#####
# X-WING ATTACK
# a game made by Tim Schick
#####

from random import random, choice, randint
from time import sleep
from collections.abc import Generator
from itertools import zip_longest
import os
from dataclasses import dataclass

STAR_PERCENTAGE = 0.06
STAR_CHARS = [c for c in "....***'''+\xB7\xB7\xB7\xB7"]
WINDOW_HEIGHT = os.get_terminal_size()[1]
PAN_DOWN_DELAY = 0.2

PLACEHOLDER_SPACE = "$"
TITLE_SCREEN_TEXT = [
    "  ██╗  ██╗     ██╗    ██╗██╗███╗   ██╗ ██████╗    ",
    "  ╚██╗██╔╝     ██║    ██║██║████╗  ██║██╔════╝    ",
    "   ╚███╔╝█████╗██║ █╗ ██║██║██╔██╗ ██║██║  ███╗   ",
    "   ██╔██╗╚════╝██║███╗██║██║██║╚██╗██║██║   ██║   ",
    "  ██╔╝ ██╗     ╚███╔███╔╝██║██║ ╚████║╚██████╔╝   ",
    "  ╚═╝  ╚═╝      ╚══╝╚══╝ ╚═╝╚═╝  ╚═══╝ ╚═════╝    ",
    "                                                  ",
    " █████╗ ████████╗████████╗ █████╗  ██████╗██╗  ██╗",
    "██╔══██╗╚══██╔══╝╚══██╔══╝██╔══██╗██╔════╝██║ ██╔╝",
    "███████║   ██║      ██║   ███████║██║     █████╔╝ ",
    "██╔══██║   ██║      ██║   ██╔══██║██║     ██╔═██╗ ",
    "██║  ██║   ██║      ██║   ██║  ██║╚██████╗██║  ██╗",
    "╚═╝  ╚═╝   ╚═╝      ╚═╝   ╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝",
    "                                                  ",
    "                 by$Tim$Schick                    "]
padding_up_down = (WINDOW_HEIGHT - len(TITLE_SCREEN_TEXT)) // 2
TITLE_SCREEN_TEXT = (padding_up_down + 2) * [""] + TITLE_SCREEN_TEXT + (padding_up_down - 1) * [""]

END_SCREEN_TEXT = [
    "████████╗██╗  ██╗███████╗ ",
    "╚══██╔══╝██║  ██║██╔════╝ ",
    "   ██║   ███████║█████╗   ",
    "   ██║   ██╔══██║██╔══╝   ",
    "   ██║   ██║  ██║███████╗ ",
    "   ╚═╝   ╚═╝  ╚═╝╚══════╝ ",
    "                          ",
    "███████╗███╗   ██╗██████╗ ",
    "██╔════╝████╗  ██║██╔══██╗",
    "█████╗  ██╔██╗ ██║██║  ██║",
    "██╔══╝  ██║╚██╗██║██║  ██║",
    "███████╗██║ ╚████║██████╔╝",
    "╚══════╝╚═╝  ╚═══╝╚═════╝ ",
    "                          ",
    "  thank$you$for$playing   "]
padding_up_down = (WINDOW_HEIGHT - len(END_SCREEN_TEXT)) // 2
END_SCREEN_TEXT = (padding_up_down + 2) * [""] + END_SCREEN_TEXT + (padding_up_down - 1) * [""]

YAVIN_TEMPLE_ASCII = [
    "                     _|_                     ",
    "                    /   \\                    ",
    "            _______/_____\\_______            ",
    "         __|___|_|_________|_|___|__         ",
    "       //    | |           | |    \\\\       ",
    "    __//_____|_|___________|_|_____\\\\__    ",
    "   //       | |             | |       \\\\   ",
    " _//________|_|_____________|_|________\\\\_ ",
    "//_________|_|_______________|_|_________\\\\"]


@dataclass
class PlayerChoice:
    choice: int
    was_valid: bool


def get_player_choice(options: list[str], default: int) -> PlayerChoice:
    options_str = "[ " + " / ".join(f"{a+1}: {b}" for a, b in enumerate(options)) + " ]\n"
    try:
        choice = int(input("What do you do? " + options_str))
        valid = True
        if choice < 1 or choice > len(options):
            choice = default
            valid = False

    except ValueError:
        choice = default
        valid = False

    print("")
    return PlayerChoice(choice, valid)


def hide_cursor():
    print("\x1b[?25l")


def show_cursor():
    print("\x1b[?25h")


def check_screen_size_valid(min_cols: int, min_rows: int) -> bool:
    cols, rows = os.get_terminal_size()
    return cols >= min_cols and rows >= min_rows


def invalid_screen_size_warning(min_cols: int, min_rows: int):
    cols, rows = os.get_terminal_size()
    print("WARNING:")
    print(f"Your terminal has a size of {cols} columns and {rows} rows,")
    print(f"but it is recommended to have at least {min_cols} columns and {min_rows} rows.")
    print("You can still play this game, but your experience will not be optimal.")
    print("Do you want to continue anyway?\n")
    choice = input("Press y to continue, or something else to quit: ")
    if choice != "y":
        quit()


def clear_screen():
    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")


def enter_next_stage():
    hide_cursor()
    input("Press Enter to continue...\n")
    show_cursor()
    clear_screen()


def gen_starline(width: int) -> str:
    def random_char() -> str:
        return choice(STAR_CHARS) if random() < STAR_PERCENTAGE else " "

    return "".join([random_char() for i in range(width)])


def create_starline_generator(width=120, height=30) -> Generator[str, None, None]:
    while True:
        yield gen_starline(width)


def merge_text(foreground: str, background: str) -> str:
    return "".join(b if f == " " else " " if f == PLACEHOLDER_SPACE else f for f, b in zip_longest(foreground, background, fillvalue=" "))


def make_text_centered(line: str, width: int) -> str:
    padding_left = (width - len(line)) // 2
    return padding_left * " " + line


def print_titlescreen(stars: Generator[str, None, None], width: int):
    clear_screen()
    full_str = "\n".join(merge_text(make_text_centered(line, width), next(stars)) for line in TITLE_SCREEN_TEXT)
    print(full_str[:-1])
    hide_cursor()
    input(merge_text("Press$Enter$to$start$", next(stars)))
    show_cursor()


def print_endscreen(stars: Generator[str, None, None], width: int):
    clear_screen()
    full_str = "\n".join(merge_text(make_text_centered(line, 80), next(stars)) for line in END_SCREEN_TEXT)
    print(full_str[:-1])
    hide_cursor()
    input(merge_text("Press$Enter$to$quit$", next(stars)))
    show_cursor()


def print_pan_down(stars: Generator[str, None, None], width: int):
    hide_cursor()
    for i in range(WINDOW_HEIGHT):
        print(next(stars))
        sleep(PAN_DOWN_DELAY)
    for i in range(WINDOW_HEIGHT - 10):
        print("")
        sleep(PAN_DOWN_DELAY)
    for temple_line in YAVIN_TEMPLE_ASCII:
        temple_line = make_text_centered(temple_line, width)
        print(temple_line)
        sleep(PAN_DOWN_DELAY)
    sleep(2)
    show_cursor()


def print_briefing():
    print("Your name is Luke Skywalker. You came into the possession of the stolen")
    print("plans for a mighty space station called the Death Star, which is able")
    print("to destroy whole planets. It is operated by an evil and oppressive Empire.")
    print("You are currently in a rebel base on the moon Yavin 4. The empire knows")
    print("where you are, and it is going to annihilate Yavin 4 to crush the rebellion.")
    print("There is, however, a way to destroy the Death Star. It requires brave pilots")
    print("to shoot proton torpedos into a tiny reactor exhaust port.")
    print("This will start a chain reaction, leading to the total destruction of the base.")
    print("You have volunteered to partake in this seemingly impossible mission.")


def play_hangar_scene() -> bool:
    print("You meet the smuggler Han Solo and his Copilot Chewbacca in the hangar.")
    print("They are about to leave. Han Solo doesn't care for the rebellion, but he")
    print("offers you to join his crew. You could make some money here, and avoid")
    print("risking your life fighting an overwhelmingly powerful opponent.")
    choice = get_player_choice(["join Han Solo", "fight the empire"], 2)

    if choice.choice == 1:
        print("You leave the rebels behind. You will sometimes wonder if you made the")
        print("correct choice. The empire's terror will get more firce,")
        print("and it's grip on the galaxy more suffocating. The Death Star")
        print("will quell all forms of resistance in their infancy.")
        print("You survive, but many others will die.")
        return False
    else:
        print("You decline.")
        print("You get to your X-wing fighter and prepare for battle.")
        print("Shortly after, you and the other pilots launch and fly to the Death Star.")
        return True


def print_early_attack():
    print("All squadrons arrive in the vicinity of the station.")
    print("One after the other enters the trench leading to the exhaust port,")
    print("and one after the other fails.")
    print("Some get shot down by turbo laser turrets, others miss the target.")
    print("You waited patiently, fighting small skirmishes against the agile")
    print("TIE fighters that launched from the Death Star.")
    print("But now it is your turn to try your luck.")
    print("You speed up and dive into the trench, heading to the exhaust port.")


def play_trench_encounter(encounter_number: int, tie_location: int | None) -> bool:
    sides = {1: "left", 3: "right"}

    if encounter_number == 1:
        print("You see a turret in the distance. You are approaching fast")
        print("and it seems to be aiming at you.")
    elif encounter_number == 2:
        print("There is another turret coming up.")
    else:
        print("A last turret is in your way.")

    if tie_location and tie_location != 2:
        print(f"Your scanners are also detecting a TIE fighter closing in on your {sides[tie_location]} side.")
        print("It will likely shoot at you, too.")
    elif tie_location == 2:
        print(f"Your scanners are also detecting a TIE fighter closing in directly behind you.")
        print("It will likely shoot at you, too.")

    side_choice = get_player_choice(["dodge left", "fly straight", "dodge right"], 2)

    player_alive = False
    if side_choice.choice == 2:
        if side_choice.was_valid:
            print("You fly straight ahead and hope that you don't get hit.")
        else:
            print("You are overwhelmed by the situation, so you fly straight ahead.")
            print("You just hope that you don't get hit.")
        print("However, your hopes are in vain.\nYou get shot down.")
    else:
        print(f"You skew to the {sides[side_choice.choice]}, narrowly avoiding the turret's fire", end="")

        if side_choice.choice == tie_location:
            print(",\nbut this positions you perfectly in the attacking TIE fighter's sight.\nIt shoots you down.")
        elif tie_location:
            print(",\nand you even avoid the TIE fighter behind you.")
            print("It fails to shoot at you and leaves the trench. Let's hope it won't come back.")
            player_alive = True
        else:
            print(".")
            player_alive = True

    return player_alive


def play_trench_run() -> bool:
    tie_encounter = randint(2, 3)

    for round in range(1, 4):
        tie_location = randint(1, 3) if round == tie_encounter else None
        if not play_trench_encounter(round, tie_location):
            return False

    return True


def print_vader_encounter():
    print("You have made it past the turrets,")
    print("so you have plenty of time to aim at the exhaust port now.")
    print("You activate your targeting computer. This is going to be easy.")
    print("Suddenly, your scanner is picking up several ships approaching from behind.")
    print("You are being followed by the powerful Darth Vader and two TIE fighters.")
    print("This is very bad! Vader is an excellent pilot and will definitely hit you,")
    print("but aborting the attack to try and loose him requires time,")
    print("which there is not much left of.")
    print("You continue your attack, hoping that you will at least get to shoot.")
    enter_next_stage()

    print("Vader fires at you and and just barely misses your ship.")
    print("Without warning, one of the TIE fighters explodes. You look around.")
    print("Han Solo and Chewbacca have returned to help!")
    print("Another shot, this time hitting the second TIE fighter. It gets thrown to")
    print("the side, touching Vader's fighter and then crashing into the Wall, exploding.")
    print("This, in turn, causes Vader to loose control of his fighter")
    print("and start spinning out into space.")


def play_targeting_choice() -> bool:
    print("You are very close to the exhaust port now. As you begin to calibrate your")
    print("targeting computer, you suddenly hear the voice of the dead Jedi Ben Kenobi:")
    print("'Use the force, Luke.'")
    choice = get_player_choice(["use the force", "use the targeting computer"], 2)

    if choice.choice == 1:
        print("You turn off your targeting computer and let the force guide you.")
        print("You take aim and fire two torpedos.")
        print("They enter the exhaust port perfectly. It's time to leave.")
        print("Shortly after, the Death Star explodes and Yavin 4 is safe. You have won.")
        return True

    else:
        if not choice.was_valid:
            print("You are confused. You saw Ben die.")
        else:
            print("You ignore the voice. You know that Ben is dead.")

        print("Either your mind is playing tricks on you, or the empire is somehow")
        print("interfering with the communication channels.")
        print("Whatever just happened: It is probably best not to listen to dead men's voices.")
        print("The targeting computer is running and has locked on to the exhaust port.")
        print("You fire two torpedoes. They miss the target and impact on the surface.")
        print("This is going to leave a mark, but it will not stop the Death Star.")
        print("There is nothing you can do but watch it vaporize Yavin 4. You have failed.")
        return False


def play_ending_ceremony():
    print("You, Han Solo and Chewbacca return to Yavin 4, where you are greeted")
    print("as heroes and awarded medals for your bravery.")


def play_game() -> bool:
    print_briefing()
    enter_next_stage()
    player_continues = play_hangar_scene()
    enter_next_stage()
    if not player_continues:
        return False

    print_early_attack()
    enter_next_stage()
    player_survives_trench = play_trench_run()
    enter_next_stage()
    if not player_survives_trench:
        return False

    print_vader_encounter()
    enter_next_stage()
    player_hits_target = play_targeting_choice()
    enter_next_stage()
    if not player_hits_target:
        return False

    play_ending_ceremony()
    enter_next_stage()
    return True


def main():
    star_generator = create_starline_generator(80, WINDOW_HEIGHT)
    print_titlescreen(star_generator, 80)
    print_pan_down(star_generator, 80)

    clear_screen()
    play_game()

    print_endscreen(star_generator, 80)


if __name__ == "__main__":
    if not check_screen_size_valid(80, 25):
        invalid_screen_size_warning(80, 25)

    try:
        main()
    # this finally block is needed to unhide the cursor in case of a keyboard interrupt
    # DO NOT REMOVE
    finally:
        show_cursor()